# -*- coding: utf-8 -*-
# This file is part of the bumpver project
# https://github.com/mbarkhau/bumpver
#
# Copyright (c) 2018-2022 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
"""Compose Regular Expressions from Patterns.

>>> version_info = PYCALVER_RE.match("v201712.0123-alpha").groupdict()
>>> assert version_info == {
...     "pycalver"    : "v201712.0123-alpha",
...     "vYYYYMM"     : "v201712",
...     "year"        : "2017",
...     "month"       : "12",
...     "build"       : ".0123",
...     "build_no"    : "0123",
...     "release"     : "-alpha",
...     "release_tag" : "alpha",
... }
>>>
>>> version_info = PYCALVER_RE.match("v201712.0033").groupdict()
>>> assert version_info == {
...     "pycalver"   : "v201712.0033",
...     "vYYYYMM"    : "v201712",
...     "year"       : "2017",
...     "month"      : "12",
...     "build"      : ".0033",
...     "build_no"   : "0033",
...     "release"    : None,
...     "release_tag": None,
... }
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import re
import typing as typ
import logging
from . import utils
from .patterns import RE_PATTERN_ESCAPES
from .patterns import Pattern
logger = logging.getLogger('bumpver.v1patterns')
PYCALVER_PATTERN = """
\\b
(?P<pycalver>
    (?P<vYYYYMM>
       v                        # "v" version prefix
       (?P<year>\\d{4})
       (?P<month>\\d{2})
    )
    (?P<build>
        \\.                      # "." build nr prefix
        (?P<build_no>\\d{4,})
    )
    (?P<release>
        \\-                      # "-" release prefix
        (?P<release_tag>alpha|beta|dev|rc|post)
    )?
)(?:\\s|$)
"""
PYCALVER_RE = re.compile(PYCALVER_PATTERN, flags=re.VERBOSE)
COMPOSITE_PART_PATTERNS = {'pep440_pycalver':
    '{year}{month}\\.{BID}(?:{pep440_tag})?', 'pycalver':
    'v{year}{month}\\.{bid}(?:-{tag})?', 'calver': 'v{year}{month}',
    'semver': '{MAJOR}\\.{MINOR}\\.{PATCH}', 'release_tag': '{tag}',
    'build': '\\.{bid}', 'release': '(?:-{tag})?', 'pep440_version':
    '{year}{month}\\.{BID}(?:{pep440_tag})?'}
PART_PATTERNS = {'year': '\\d{4}', 'month': '(?:0[0-9]|1[0-2])',
    'month_short': '(?:1[0-2]|[1-9])', 'build_no': '\\d{4,}', 'pep440_tag':
    '(?:a|b|dev|rc|post)?\\d*', 'tag': '(?:alpha|beta|dev|rc|post|final)',
    'yy': '\\d{2}', 'yyyy': '\\d{4}', 'quarter': '[1-4]', 'iso_week':
    '(?:[0-4]\\d|5[0-3])', 'us_week': '(?:[0-4]\\d|5[0-3])', 'dom':
    '(0[1-9]|[1-2][0-9]|3[0-1])', 'dom_short': '([1-9]|[1-2][0-9]|3[0-1])',
    'doy': '(?:[0-2]\\d\\d|3[0-5][0-9]|36[0-6])', 'doy_short':
    '(?:[0-2]\\d\\d|3[0-5][0-9]|36[0-6])', 'MAJOR': '\\d+', 'MINOR': '\\d+',
    'MM': '\\d{2,}', 'MMM': '\\d{3,}', 'MMMM': '\\d{4,}', 'MMMMM':
    '\\d{5,}', 'PATCH': '\\d+', 'PP': '\\d{2,}', 'PPP': '\\d{3,}', 'PPPP':
    '\\d{4,}', 'PPPPP': '\\d{5,}', 'bid': '\\d{4,}', 'BID': '[1-9]\\d*',
    'BB': '[1-9]\\d{1,}', 'BBB': '[1-9]\\d{2,}', 'BBBB': '[1-9]\\d{3,}',
    'BBBBB': '[1-9]\\d{4,}', 'BBBBBB': '[1-9]\\d{5,}', 'BBBBBBB':
    '[1-9]\\d{6,}'}
PATTERN_PART_FIELDS = {'year': 'year', 'month': 'month', 'month_short':
    'month', 'pep440_tag': 'tag', 'tag': 'tag', 'yy': 'year', 'yyyy':
    'year', 'quarter': 'quarter', 'iso_week': 'iso_week', 'us_week':
    'us_week', 'dom': 'dom', 'doy': 'doy', 'dom_short': 'dom', 'doy_short':
    'doy', 'MAJOR': 'major', 'MINOR': 'minor', 'MM': 'minor', 'MMM':
    'minor', 'MMMM': 'minor', 'MMMMM': 'minor', 'PP': 'patch', 'PPP':
    'patch', 'PPPP': 'patch', 'PPPPP': 'patch', 'PATCH': 'patch',
    'build_no': 'bid', 'bid': 'bid', 'BID': 'bid', 'BB': 'bid', 'BBB':
    'bid', 'BBBB': 'bid', 'BBBBB': 'bid', 'BBBBBB': 'bid', 'BBBBBBB': 'bid'}
FULL_PART_FORMATS = {'pep440_pycalver':
    '{year}{month:02}.{BID}{pep440_tag}', 'pycalver':
    'v{year}{month:02}.{bid}{release}', 'calver': 'v{year}{month:02}',
    'semver': '{MAJOR}.{MINOR}.{PATCH}', 'release_tag': '{tag}', 'build':
    '.{bid}', 'month': '{month:02}', 'month_short': '{month}', 'build_no':
    '{bid}', 'iso_week': '{iso_week:02}', 'us_week': '{us_week:02}', 'dom':
    '{dom:02}', 'doy': '{doy:03}', 'dom_short': '{dom}', 'doy_short':
    '{doy}', 'pep440_version': '{year}{month:02}.{BID}{pep440_tag}',
    'version': 'v{year}{month:02}.{bid}{release}'}


def _replace_pattern_parts(pattern):
    for part_name, part_pattern in PART_PATTERNS.items():
        named_part_pattern = '(?P<{0}>{1})'.format(part_name, part_pattern)
        placeholder = '\\{' + part_name + '\\}'
        pattern = pattern.replace(placeholder, named_part_pattern)
    return pattern


def _init_composite_patterns():
    for part_name, part_pattern in COMPOSITE_PART_PATTERNS.items():
        part_pattern = part_pattern.replace('{', '\\{').replace('}', '\\}')
        pattern_str = _replace_pattern_parts(part_pattern)
        PART_PATTERNS[part_name] = pattern_str


_init_composite_patterns()


def _compile_pattern_re(normalized_pattern):
    escaped_pattern = normalized_pattern
    for char, escaped in RE_PATTERN_ESCAPES:
        escaped_pattern = escaped_pattern.replace(char, escaped)
    pattern_str = _replace_pattern_parts(escaped_pattern)
    return re.compile(pattern_str)


def _normalized_pattern(version_pattern, raw_pattern):
    res = raw_pattern.replace('{version}', version_pattern)
    if version_pattern == '{pycalver}':
        res = res.replace('{pep440_version}', '{pep440_pycalver}')
    elif version_pattern == '{semver}':
        res = res.replace('{pep440_version}', '{semver}')
    elif version_pattern == 'v{year}{month}{build}{release}':
        res = res.replace('{pep440_version}', '{year}{month}.{BID}{pep440_tag}'
            )
    elif version_pattern == '{year}{month}{build}{release}':
        res = res.replace('{pep440_version}', '{year}{month}.{BID}{pep440_tag}'
            )
    elif version_pattern == 'v{year}{build}{release}':
        res = res.replace('{pep440_version}', '{year}.{BID}{pep440_tag}')
    elif version_pattern == '{year}{build}{release}':
        res = res.replace('{pep440_version}', '{year}.{BID}{pep440_tag}')
    elif '{pep440_version}' in raw_pattern:
        logger.warning("No mapping of '{0}' to '{pep440_version}'".format(
            version_pattern))
    return res


@utils.memo
def compile_pattern(version_pattern, raw_pattern=None):
    _raw_pattern = version_pattern if raw_pattern is None else raw_pattern
    normalized_pattern = _normalized_pattern(version_pattern, _raw_pattern)
    regexp = _compile_pattern_re(normalized_pattern)
    return Pattern(version_pattern, normalized_pattern, regexp)


def compile_patterns(version_pattern, raw_patterns):
    return [compile_pattern(version_pattern, raw_pattern) for raw_pattern in
        raw_patterns]



# -*- coding: utf-8 -*-
# This file is part of the bumpver project
# https://github.com/mbarkhau/bumpver
#
# Copyright (c) 2018-2022 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import typing as typ
try:
    import builtins
except ImportError:
    import __builtin__ as builtins
import functools
str = getattr(builtins, 'unicode', str)


def memo(func):
    cache = {}

    @functools.wraps(func)
    def wrapper(*args):
        key = str(args)
        if key not in cache:
            cache[key] = func(*args)
        return cache[key]
    return wrapper



from communicationApp.telegram import Telegram
from typing import  Any


def createClient(app_name: str, **kwargs: Any):
  try:
    app_name = app_name.capitalize()
    klass = globals()[app_name]
    print(klass)
    return klass(**kwargs)
  except KeyError:
    return {"status_code": 404, "error": f"{app_name} is not support"}




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the bumpver project
# https://github.com/mbarkhau/bumpver
#
# Copyright (c) 2018-2022 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
"""cli module for BumpVer."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import io
import re
import sys
import typing as typ
import logging
import datetime as dt
import subprocess as sp
import click
try:
    import builtins
except ImportError:
    import __builtin__ as builtins
import colorama
str = getattr(builtins, 'unicode', str)
from . import vcs
from . import config
from . import rewrite
from . import version
from . import patterns
from . import regexfmt
from . import v1rewrite
from . import v1version
from . import v2rewrite
from . import v2version
from . import v1patterns
from . import v2patterns
try:
    import pretty_traceback
    pretty_traceback.install()
except ImportError:
    pass
click.disable_unicode_literals_warning = True
logger = logging.getLogger('bumpver.cli')
_VERBOSE = 0


def _configure_logging(verbose=0):
    global _VERBOSE
    _VERBOSE = verbose
    if verbose >= 2:
        log_format = (
            '%(asctime)s.%(msecs)03d %(levelname)-7s %(name)-17s - %(message)s'
            )
        log_level = logging.DEBUG
    elif verbose == 1:
        log_format = '%(levelname)-7s - %(message)s'
        log_level = logging.INFO
    else:
        log_format = '%(levelname)-7s - %(message)s'
        log_level = logging.INFO
    logging.basicConfig(level=log_level, format=log_format, datefmt=
        '%Y-%m-%dT%H:%M:%S')
    logger.debug('Logging configured.')


VALID_RELEASE_TAG_VALUES = 'alpha', 'beta', 'rc', 'post', 'final'
_current_date = dt.date.today().isoformat()


def _validate_date(date, pin_date):
    if date and pin_date:
        logger.error(
            "Can only use either --pin-date or --date='{0}', not both.".
            format(date))
        sys.exit(1)
    if date is None:
        return None
    try:
        dt_val = dt.datetime.strptime(date, '%Y-%m-%d')
        return dt_val.date()
    except ValueError:
        logger.error(
            "Invalid parameter --date='{0}', must match format YYYY-0M-0D."
            .format(date), exc_info=True)
        sys.exit(1)


def _validate_release_tag(tag):
    if tag is None:
        return
    if tag in VALID_RELEASE_TAG_VALUES:
        return
    logger.error('Invalid argument --tag={0}'.format(tag))
    logger.error('Valid arguments are: {0}'.format(', '.join(
        VALID_RELEASE_TAG_VALUES)))
    sys.exit(1)


def _validate_flags(raw_pattern, major, minor, patch):
    if '{' in raw_pattern and '}' in raw_pattern:
        return
    valid = True
    if major and 'MAJOR' not in raw_pattern:
        logger.error("Flag --major is not applicable to pattern '{0}'".
            format(raw_pattern))
        valid = False
    if minor and 'MINOR' not in raw_pattern:
        logger.error("Flag --minor is not applicable to pattern '{0}'".
            format(raw_pattern))
        valid = False
    if patch and 'PATCH' not in raw_pattern:
        logger.error("Flag --patch is not applicable to pattern '{0}'".
            format(raw_pattern))
        valid = False
    if not valid:
        sys.exit(1)


def _log_no_change(subcmd, version_pattern):
    is_semver = ('{semver}' in version_pattern or 'MAJOR' in
        version_pattern and 'MAJOR' in version_pattern and 'PATCH' in
        version_pattern)
    if is_semver:
        logger.warning(
            'bumpver {0} [--major/--minor/--patch] required for use with SemVer.'
            .format(subcmd))
    else:
        available_flags = [('--' + part.lower()) for part in ['MAJOR',
            'MINOR', 'PATCH'] if part in version_pattern]
        if available_flags:
            available_flags_str = '/'.join(available_flags)
            logger.info('Perhaps try: bumpver {0} {1} '.format(subcmd,
                available_flags_str))


def _get_normalized_pattern(raw_pattern, version_pattern):
    is_version_pattern_required = ('{version}' in raw_pattern or 
        '{pep440_version}' in raw_pattern)
    if is_version_pattern_required and version_pattern is None:
        logger.error(
            'Argument --version-pattern=<PATTERN> is required for placeholders: {version}/{pep440_version}.'
            )
        sys.exit(1)
    elif version_pattern is None:
        _version_pattern = 'INVALID'
    else:
        _version_pattern = version_pattern
    if is_version_pattern_required:
        return v2patterns.normalize_pattern(_version_pattern, raw_pattern)
    else:
        return raw_pattern


verbose_option = click.option('-v', '--verbose', count=True, help=
    'Control log level. -vv for debug level.')
dry_option = click.option('-d', '--dry', default=False, is_flag=True, help=
    "Display diff of changes, don't rewrite files.")
fetch_option = click.option('-f/-n', '--fetch/--no-fetch', is_flag=True,
    default=True, help='Sync tags from remote origin.')
env_option = click.option('-e', '--env', is_flag=True, default=False, help=
    'Print version state for use with shell scripts: eval $(bumpver show --env)'
    )


def version_options(function):
    decorators = [click.option('--major', is_flag=True, default=False, help
        ='Increment MAJOR component.'), click.option('-m', '--minor',
        is_flag=True, default=False, help='Increment MINOR component.'),
        click.option('-p', '--patch', is_flag=True, default=False, help=
        'Increment PATCH component.'), click.option('-t', '--tag', default=
        None, metavar='<NAME>', help=
        'Override release tag of current_version. Valid options are: {0}.'.
        format(', '.join(VALID_RELEASE_TAG_VALUES))), click.option(
        '--tag-num', is_flag=True, default=False, help=
        'Increment release tag number (rc1, rc2, rc3..).'), click.option(
        '--pin-date', is_flag=True, default=False, help=
        'Leave date components unchanged.'), click.option('--date', default
        =None, metavar='<ISODATE>', help=
        'Set explicit date in format YYYY-0M-0D (e.g. {0}).'.format(
        _current_date)), click.option('--set-version', default=None,
        metavar='<VERSION>', help='Set version explicitly.')]
    decorated = function
    for decorator in decorators:
        decorated = decorator(decorated)
    return decorated


@click.group()
@click.version_option(version='2022.1116')
@click.help_option()
@verbose_option
def cli(verbose=0):
    """Automatically update version strings in plaintext files."""
    if verbose:
        _configure_logging(verbose=max(_VERBOSE, verbose))


@cli.command()
@click.argument('old_version')
@click.argument('pattern')
@verbose_option
@version_options
def test(old_version, pattern, verbose=0, major=False, minor=False, patch=
    False, tag=None, tag_num=False, pin_date=False, date=None, set_version=None
    ):
    """Increment a version number for demo purposes."""
    _configure_logging(verbose=max(_VERBOSE, verbose))
    _validate_release_tag(tag)
    raw_pattern = pattern
    _validate_flags(raw_pattern, major, minor, patch)
    maybe_date = _validate_date(date, pin_date)
    if set_version is None:
        new_version = incr_dispatch(old_version, raw_pattern=raw_pattern,
            major=major, minor=minor, patch=patch, tag=tag, tag_num=tag_num,
            pin_date=pin_date, maybe_date=maybe_date)
    else:
        new_version = set_version
    if new_version is None:
        _log_no_change('test', raw_pattern)
        sys.exit(1)
    if not _is_valid_version(raw_pattern, old_version, new_version):
        if set_version:
            logger.error("Invalid argument --set-version='{0}'".format(
                set_version))
        sys.exit(1)
    pep440_version = version.to_pep440(new_version)
    click.echo('New Version: {0}'.format(new_version))
    if new_version != pep440_version:
        click.echo('PEP440     : {0}'.format(pep440_version))


def _grep_text(pattern, text, color):
    all_lines = text.splitlines()
    for match in pattern.regexp.finditer(text):
        match_start, match_end = match.span()
        line_idx = text[:match_start].count('\n')
        line_start = text.rfind('\n', 0, match_start) + 1
        line_end = text.find('\n', match_end, -1)
        if color:
            matched_line = text[line_start:match_start
                ] + colorama.Style.BRIGHT + text[match_start:match_end
                ] + colorama.Style.RESET_ALL + text[match_end:line_end]
        else:
            matched_line = text[line_start:match_start] + text[match_start:
                match_end] + text[match_end:line_end]
        lines_offset = max(0, line_idx - 1) + 1
        lines = all_lines[line_idx - 1:line_idx + 2]
        if line_idx == 0:
            lines[0] = matched_line
        else:
            lines[1] = matched_line
        prefixed_lines = ['{0:>4}: {1}'.format(lines_offset + i, line) for 
            i, line in enumerate(lines)]
        yield '\n'.join(prefixed_lines)


def _grep(raw_pattern, file_ios, color):
    pattern = v2patterns.compile_pattern(raw_pattern)
    match_count = 0
    for file_io in file_ios:
        text = file_io.read()
        match_strs = list(_grep_text(pattern, text, color))
        if len(match_strs) > 0:
            if len(file_ios) > 1:
                print(file_io.name)
            for match_str in match_strs:
                print(match_str)
            print()
        match_count += len(match_strs)
    if match_count == 0:
        logger.error("Pattern not found: '{0}'".format(raw_pattern))
    if match_count == 0 or _VERBOSE:
        pyexpr_regex = regexfmt.pyexpr_regex(pattern.regexp.pattern)
        print('# ' + regexfmt.regex101_url(pattern.regexp.pattern))
        print(pyexpr_regex)
        print()
    if match_count == 0:
        sys.exit(1)


@cli.command()
@verbose_option
@click.option('--version-pattern', default=None, metavar='<PATTERN>', help=
    'Pattern to use for placeholders: {version}/{pep440_version}')
@click.argument('pattern')
@click.argument('files', nargs=-1, type=click.File('r'))
def grep(pattern, files, version_pattern=None, verbose=0):
    """Search file(s) for a version pattern."""
    verbose = max(_VERBOSE, verbose)
    _configure_logging(verbose)
    raw_pattern = pattern
    normalized_pattern = _get_normalized_pattern(raw_pattern, version_pattern)
    isatty = getattr(sys.stdout, 'isatty', lambda : False)
    if isatty():
        colorama.init()
        try:
            _grep(normalized_pattern, files, color=True)
        finally:
            colorama.deinit()
    else:
        _grep(normalized_pattern, files, color=False)


@cli.command()
@verbose_option
@fetch_option
@env_option
def show(verbose=0, fetch=True, env=False):
    """Show current version of your project."""
    _configure_logging(verbose=max(_VERBOSE, verbose))
    _, cfg = config.init(project_path='.')
    if cfg is None:
        logger.error(
            "Could not parse configuration. Perhaps try 'bumpver init'.")
        sys.exit(1)
    cfg = _update_cfg_from_vcs(cfg, fetch)
    if env:
        version_info = v2version.parse_version_info(cfg.current_version,
            cfg.version_pattern)
        for key, val in version_info._asdict().items():
            click.echo('{0}={1}'.format(key.upper(), val if val else ''))
        click.echo('CURRENT_VERSION={0}'.format(cfg.current_version))
        click.echo('PEP440_VERSION={0}'.format(cfg.pep440_version))
    else:
        click.echo('Current Version: {0}'.format(cfg.current_version))
        click.echo('PEP440         : {0}'.format(cfg.pep440_version))


def _colored_diff_lines(diff):
    for line in diff.splitlines():
        if line.startswith('+++') or line.startswith('---'):
            yield line
        elif line.startswith('+'):
            yield '\x1b[32m' + line + '\x1b[0m'
        elif line.startswith('-'):
            yield '\x1b[31m' + line + '\x1b[0m'
        elif line.startswith('@'):
            yield '\x1b[36m' + line + '\x1b[0m'
        else:
            yield line


def _v2_get_diff(cfg, new_version):
    old_vinfo = v2version.parse_version_info(cfg.current_version, cfg.
        version_pattern)
    new_vinfo = v2version.parse_version_info(new_version, cfg.version_pattern)
    return v2rewrite.diff(old_vinfo, new_vinfo, cfg.file_patterns)


def _v1_get_diff(cfg, new_version):
    old_vinfo = v1version.parse_version_info(cfg.current_version, cfg.
        version_pattern)
    new_vinfo = v1version.parse_version_info(new_version, cfg.version_pattern)
    return v1rewrite.diff(old_vinfo, new_vinfo, cfg.file_patterns)


def get_diff(cfg, new_version):
    if cfg.is_new_pattern:
        return _v2_get_diff(cfg, new_version)
    else:
        return _v1_get_diff(cfg, new_version)


def _print_diff_str(diff):
    colored_diff = '\n'.join(_colored_diff_lines(diff))
    if sys.stdout.isatty():
        click.echo(colored_diff)
    else:
        click.echo(diff)


def _print_diff(cfg, new_version):
    try:
        diff = get_diff(cfg, new_version)
        _print_diff_str(diff)
    except OSError as err:
        logger.error(str(err))
        sys.exit(1)
    except rewrite.NoPatternMatch as ex:
        logger.error(str(ex))
        sys.exit(1)


def _is_valid_version(raw_pattern, old_version, new_version):
    is_new_pattern = '{' not in raw_pattern and '}' not in raw_pattern
    try:
        if is_new_pattern:
            v2version.parse_version_info(new_version, raw_pattern)
        else:
            v1version.parse_version_info(new_version, raw_pattern)
    except version.PatternError:
        logger.error("Invalid version '{0}' for pattern '{1}'".format(
            new_version, raw_pattern))
        return False
    if version.parse_version(new_version) <= version.parse_version(old_version
        ):
        logger.error(
            'Invariant violated: New version must be greater than old version '
            )
        logger.error("  Failed Invariant: '{0}' > '{1}'".format(new_version,
            old_version))
        return False
    else:
        return True


def incr_dispatch(old_version, raw_pattern, **kwargs):
    major = kwargs.get('major', False)
    minor = kwargs.get('minor', False)
    patch = kwargs.get('patch', False)
    tag = kwargs.get('tag', None)
    tag_num = kwargs.get('tag_num', False)
    pin_date = kwargs.get('pin_date', False)
    maybe_date = kwargs.get('maybe_date', None)
    v1_parts = list(v1patterns.PART_PATTERNS) + list(v1patterns.
        FULL_PART_FORMATS)
    has_v1_part = any('{' + part + '}' in raw_pattern for part in v1_parts)
    if _VERBOSE:
        if has_v1_part:
            pattern = v1patterns.compile_pattern(raw_pattern)
        else:
            pattern = v2patterns.compile_pattern(raw_pattern)
        logger.info('Using pattern ' + raw_pattern)
        logger.info('regex = ' + regexfmt.pyexpr_regex(pattern.regexp.pattern))
    if has_v1_part:
        return v1version.incr(old_version, raw_pattern=raw_pattern, major=
            major, minor=minor, patch=patch, tag=tag, tag_num=tag_num,
            pin_date=pin_date, maybe_date=maybe_date)
    else:
        return v2version.incr(old_version, raw_pattern=raw_pattern, major=
            major, minor=minor, patch=patch, tag=tag, tag_num=tag_num,
            pin_date=pin_date, maybe_date=maybe_date)


def _update(cfg, new_version, commit_message, allow_dirty=False):
    vcs_api = None
    if cfg.commit:
        try:
            vcs_api = vcs.get_vcs_api()
        except OSError:
            logger.warning('Version Control System not found, skipping commit.'
                )
    filepaths = set(cfg.file_patterns.keys())
    if vcs_api:
        vcs.assert_not_dirty(vcs_api, filepaths, allow_dirty)
    try:
        if cfg.is_new_pattern:
            new_v2_vinfo = v2version.parse_version_info(new_version, cfg.
                version_pattern)
            v2rewrite.rewrite_files(cfg.file_patterns, new_v2_vinfo)
        else:
            new_v1_vinfo = v1version.parse_version_info(new_version, cfg.
                version_pattern)
            v1rewrite.rewrite_files(cfg.file_patterns, new_v1_vinfo)
    except rewrite.NoPatternMatch as ex:
        logger.error(str(ex))
        sys.exit(1)
    if vcs_api:
        vcs.commit(cfg, vcs_api, filepaths, new_version, commit_message)


def _try_update(cfg, new_version, commit_message, allow_dirty=False):
    try:
        _update(cfg, new_version, commit_message, allow_dirty)
    except sp.CalledProcessError as ex:
        logger.error('Error running subcommand: {0}'.format(ex.cmd))
        if ex.stdout:
            sys.stdout.write(ex.stdout.decode('utf-8'))
        if ex.stderr:
            sys.stderr.write(ex.stderr.decode('utf-8'))
        sys.exit(1)


@cli.command()
@verbose_option
@dry_option
def init(verbose=0, dry=False):
    """Initialize [bumpver] configuration."""
    _configure_logging(verbose=max(_VERBOSE, verbose))
    ctx, cfg = config.init(project_path='.', cfg_missing_ok=True)
    if cfg:
        logger.error('Configuration already initialized in {0}'.format(ctx.
            config_rel_path))
        sys.exit(1)
    if dry:
        click.echo("Exiting because of '-d/--dry'. Would have written to {0}:"
            .format(ctx.config_rel_path))
        cfg_text = config.default_config(ctx)
        click.echo('\n    ' + '\n    '.join(cfg_text.splitlines()))
        sys.exit(0)
    config.write_content(ctx)


def get_latest_vcs_version_tag(cfg, fetch):
    all_tags = vcs.get_tags(fetch=fetch)
    if cfg.is_new_pattern:
        version_tags = [tag for tag in all_tags if v2version.is_valid(tag,
            cfg.version_pattern)]
    else:
        version_tags = [tag for tag in all_tags if v1version.is_valid(tag,
            cfg.version_pattern)]
    if version_tags:
        version_tags.sort(key=version.parse_version, reverse=True)
        _debug_tags = ', '.join(version_tags[:3])
        logger.debug('found tags: {0} ... ({1} in total)'.format(
            _debug_tags, len(version_tags)))
        return version_tags[0]
    else:
        return None


def _update_cfg_from_vcs(cfg, fetch):
    latest_version_tag = get_latest_vcs_version_tag(cfg, fetch)
    if latest_version_tag is None:
        logger.debug('no vcs tags found')
        return cfg
    else:
        latest_version_pep440 = version.to_pep440(latest_version_tag)
        if version.parse_version(latest_version_tag) <= version.parse_version(
            cfg.current_version):
            return cfg
        else:
            logger.info('Working dir version        : {0}'.format(cfg.
                current_version))
            logger.info('Latest version from VCS tag: {0}'.format(
                latest_version_tag))
            return cfg._replace(current_version=latest_version_tag,
                pep440_version=latest_version_pep440)


def _parse_vcs_options(cfg, commit=None, tag_commit=None, push=None):
    if commit is False and tag_commit:
        raise ValueError(
            '--no-commit and --tag-commit cannot be used at the same time')
    if commit is False and push:
        raise ValueError(
            '--no-commit and --push cannot be used at the same time')
    if commit is not None:
        cfg = cfg._replace(commit=commit)
    if not cfg.commit and tag_commit:
        raise ValueError(
            '--tag-commit requires either --commit or commit=True in your config'
            )
    if not cfg.commit and push:
        raise ValueError(
            '--push requires either --commit or commit=True in your config')
    if tag_commit is not None:
        cfg = cfg._replace(tag=tag_commit)
    if push is not None:
        cfg = cfg._replace(push=push)
    return cfg


@cli.command()
@dry_option
@fetch_option
@verbose_option
@click.option('--allow-dirty', default=False, is_flag=True, help=
    'Commit even when working directory is has uncomitted changes. (WARNING: The commit will still be aborted if there are uncomitted to files with version strings.'
    )
@version_options
@click.option('-c', '--commit-message', default=None, metavar='<TMPL>',
    help='Set commit message template.')
@click.option('--commit/--no-commit', default=None, help=
    'Create a commit with all updated files.')
@click.option('--tag-commit/--no-tag-commit', default=None, help=
    'Tag the newly created commit.')
@click.option('--push/--no-push', default=None, help=
    'Push to the default remote.')
def update(dry=False, allow_dirty=False, fetch=True, verbose=0, major=False,
    minor=False, patch=False, tag=None, tag_num=False, pin_date=False, date
    =None, set_version=None, commit_message=None, commit=None, tag_commit=
    None, push=None):
    """Update project files with the incremented version string."""
    verbose = max(_VERBOSE, verbose)
    _configure_logging(verbose)
    _validate_release_tag(tag)
    maybe_date = _validate_date(date, pin_date)
    _, cfg = config.init(project_path='.')
    if cfg is None:
        logger.error('Could not parse configuration.')
        sys.exit(1)
    try:
        cfg = _parse_vcs_options(cfg, commit, tag_commit, push)
    except ValueError as ex:
        logger.warning('Invalid argument: {0}'.format(ex))
        sys.exit(1)
    cfg = _update_cfg_from_vcs(cfg, fetch)
    old_version = cfg.current_version
    if set_version is None:
        new_version = incr_dispatch(old_version, raw_pattern=cfg.
            version_pattern, major=major, minor=minor, patch=patch, tag=tag,
            tag_num=tag_num, pin_date=pin_date, maybe_date=maybe_date)
    else:
        new_version = set_version
    if new_version is None:
        _log_no_change('update', cfg.version_pattern)
        sys.exit(1)
    if not _is_valid_version(cfg.version_pattern, old_version, new_version):
        if set_version:
            logger.error("Invalid argument --set-version='{0}'".format(
                set_version))
        sys.exit(1)
    logger.info('Old Version: {0}'.format(old_version))
    logger.info('New Version: {0}'.format(new_version))
    if dry or verbose >= 2:
        _print_diff(cfg, new_version)
    if commit_message is None:
        commit_msg_template = cfg.commit_message
    else:
        commit_msg_template, _ = re.subn('\\b(OLD|NEW)\\b', '{\\1_VERSION}',
            commit_message)
    commit_message_kwargs = {'new_version': new_version, 'old_version':
        old_version, 'NEW_VERSION': new_version, 'OLD_VERSION': old_version,
        'new_version_pep440': version.to_pep440(new_version),
        'old_version_pep440': version.to_pep440(old_version)}
    try_commit_message = commit_msg_template.format(**commit_message_kwargs)
    if dry:
        return
    _try_update(cfg, new_version, try_commit_message, allow_dirty)


if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
# This file is part of the bumpver project
# https://github.com/mbarkhau/bumpver
#
# Copyright (c) 2018-2022 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import sys
try:
    import builtins
except ImportError:
    import __builtin__ as builtins
import typing as typ
str = getattr(builtins, 'unicode', str)
PY2 = sys.version_info.major < 3
try:
    try:
        from urllib.parse import quote as py3_stdlib_quote
    except ImportError:
        from urlparse import quote as py3_stdlib_quote
except ImportError:
    from urllib import quote as py2_stdlib_quote


def quote(string, safe='/', encoding=None, errors=None):
    if not isinstance(string, str):
        errmsg = 'Expected str/unicode but got {0}'.format(type(string))
        raise TypeError(errmsg)
    if encoding is None:
        _encoding = 'utf-8'
    else:
        _encoding = encoding
    if errors is None:
        _errors = 'strict'
    else:
        _errors = errors
    if PY2:
        data = string.encode(_encoding)
        res = py2_stdlib_quote(data, safe=safe.encode(_encoding))
        return res.decode(_encoding, errors=_errors)
    else:
        return py3_stdlib_quote(string, safe=safe, encoding=_encoding,
            errors=_errors)



from abc import ABC, abstractmethod

class CommunicationApp(ABC):

  def __init__(self, app):
    self.app = app

  @abstractmethod
  def send_message(self, text):
    pass
  @abstractmethod
  def get_response(self):
    pass



import re
import requests
import time


class TeleBot(object):

    def __init__(self, import_name):
        self.import_name = import_name
        self.update_rules = list()
        self.config = dict(
            api_key=None,
            requests_kwargs=dict(
                timeout=60,
            ),
        )
        self.offset = 0
        self.whoami = None

    def add_update_rule(self, rule, endpoint=None, view_func=None, **options):
        self.update_rules.append(dict(
            rule=re.compile(rule),
            endpoint=endpoint,
            view_func=view_func,
            options=dict(**options),
        ))

    def route(self, rule, **options):
        """A decorator that is used to register a view function for a
        given URL rule.  This does the same thing as :meth:`add_url_rule`
        but is intended for decorator usage::
            @app.route('/')
            def index():
                return 'Hello World'
        For more information refer to :ref:`url-route-registrations`.
        :param rule: the URL rule as string
        :param endpoint: the endpoint for the registered URL rule.  Flask
                         itself assumes the name of the view function as
                         endpoint
        :param options: the options to be forwarded to the underlying
                        :class:`~werkzeug.routing.Rule` object.  A change
                        to Werkzeug is handling of method options.  methods
                        is a list of methods this rule should be limited
                        to (``GET``, ``POST`` etc.).  By default a rule
                        just listens for ``GET`` (and implicitly ``HEAD``).
                        Starting with Flask 0.6, ``OPTIONS`` is implicitly
                        added and handled by the standard request handling.
        """
        def decorator(f):
            endpoint = options.pop('endpoint', None)
            self.add_update_rule(rule, endpoint, f, **options)
            return f
        return decorator

    def process_update(self, update):
        self.offset = max(self.offset, update.get('update_id', 0)) + 1

        for x in self.update_rules:
            # TODO: Find a good pattern to detect each type and process
            #       accordingly.
            if 'message' in update and 'text' in update['message'] and \
                    x['rule'].match(update['message']['text']):
                m = x['rule'].match(update['message']['text'])
                x['view_func'](update['message'],
                               *m.groups(),
                               **m.groupdict())

    def process_updates(self, updates):
        if updates.get('ok', False) is True:
            for msg in updates['result']:
                self.process_update(msg)

    def _start(self):
        '''Requests bot information based on current api_key, and sets
        self.whoami to dictionary with username, first_name, and id of the
        configured bot.

        '''
        if self.whoami is None:
            me = self.get_me()
            if me.get('ok', False):
                self.whoami = me['result']
            else:
                raise ValueError("Bot Cannot request information, check "
                                 "api_key")

    def poll(self, offset=None, poll_timeout=600, cooldown=60, debug=False):
        '''These should also be in the config section, but some here for
        overrides

        '''
        if self.config['api_key'] is None:
            raise ValueError("config api_key is undefined")

        if offset or self.config.get('offset', None):
            self.offset = offset or self.config.get('offset', None)

        self._start()

        while True:
            try:
                response = self.get_updates(poll_timeout, self.offset)
                if response.get('ok', False) is False:
                    raise ValueError(response['error'])
                else:
                    self.process_updates(response)
            except Exception as e:
                print("Error: Unknown Exception")
                print(e)
                if debug:
                    raise e
                else:
                    time.sleep(cooldown)

    def listen(self):
        raise NotImplemented

    def _bot_cmd(self, method, endpoint, *args, **kwargs):
        base_api = "https://api.telegram.org/bot{api_key}/{endpoint}"
        endpoint = base_api.format(api_key=self.config['api_key'],
                                   endpoint=endpoint)

        try:
            response = method(endpoint,
                              data=kwargs.get('data', None),
                              params=kwargs.get('params', {}),
                              **self.config['requests_kwargs'])

            if response.status_code != 200:
                raise ValueError('Got unexpected response. ({}) - {}'.
                                 format(response.status_code, response.text))

            return response.json()
        except Exception as e:
            return {
                'ok': False,
                'error': str(e),
            }

    def get_me(self):
        '''A simple method for testing your bot's auth token. Requires no
        parameters. Returns basic information about the bot in form of a `User
        object.

        '''
        return self._bot_cmd(requests.get, 'getMe')

    def send_message(self, chat_id, text):
        data = dict(
            chat_id=chat_id,
            text=text,
        )

        return self._bot_cmd(requests.post, 'sendMessage', data=data)

    def forward_message(self):
        raise NotImplemented("forward_message needs work")

    def send_photo(self):
        raise NotImplemented("send_photo needs work")

    def send_audio(self):
        raise NotImplemented("send_audio needs work")

    def send_document(self):
        raise NotImplemented("send_document needs work")

    def send_sticker(self):
        raise NotImplemented("send_sticker needs work")

    def send_video(self):
        raise NotImplemented("send_video needs work")

    def send_location(self):
        raise NotImplemented("send_location needs work")

    def send_chat_action(self):
        raise NotImplemented("send_chat_action needs work")

    def get_user_profile_photos(self):
        raise NotImplemented("get_user_profile_photos needs work")

    def get_updates(self, timeout=0, offset=None):
        params = dict(
            timeout=timeout,
            offset=offset,
        )
        return self._bot_cmd(requests.get, 'getUpdates', params=params)

    def set_webhook(self):
        raise NotImplemented("set_webhook needs work")



# -*- coding: utf-8 -*-
# This file is part of the bumpver project
# https://github.com/mbarkhau/bumpver
#
# Copyright (c) 2018-2022 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
"""BumpVer: A CLI program for versioning."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
__version__ = '2022.1116'



# -*- coding: utf-8 -*-
# This file is part of the bumpver project
# https://github.com/mbarkhau/bumpver
#
# Copyright (c) 2018-2022 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
"""Compose Regular Expressions from Patterns.

>>> pattern = compile_pattern("vYYYY0M.BUILD[-TAG]")
>>> version_info = pattern.regexp.match("v201712.0123-alpha")
>>> assert version_info.groupdict() == {
...     "year_y" : "2017",
...     "month"  : "12",
...     "bid"    : "0123",
...     "tag"    : "alpha",
... }
>>>
>>> version_info = pattern.regexp.match("201712.1234")
>>> assert version_info is None

>>> version_info = pattern.regexp.match("v201713.1234")
>>> assert version_info is None

>>> version_info = pattern.regexp.match("v201712.1234")
>>> assert version_info.groupdict() == {
...     "year_y" : "2017",
...     "month"  : "12",
...     "bid"    : "1234",
...     "tag"    : None,
... }
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import re
try:
    import builtins
except ImportError:
    import __builtin__ as builtins
import typing as typ
str = getattr(builtins, 'unicode', str)
from . import utils
from .patterns import RE_PATTERN_ESCAPES
from .patterns import Pattern
PART_PATTERNS = {'YYYY': '[1-9][0-9]{3}', 'YY': '[1-9][0-9]?', '0Y':
    '[0-9]{2}', 'GGGG': '[1-9][0-9]{3}', 'GG': '[1-9][0-9]?', '0G':
    '[0-9]{2}', 'Q': '[1-4]', 'MM': '1[0-2]|[1-9]', '0M': '1[0-2]|0[1-9]',
    'DD': '3[0-1]|[1-2][0-9]|[1-9]', '0D': '3[0-1]|[1-2][0-9]|0[1-9]',
    'JJJ': '36[0-6]|3[0-5][0-9]|[1-2][0-9][0-9]|[1-9][0-9]|[1-9]', '00J':
    '36[0-6]|3[0-5][0-9]|[1-2][0-9][0-9]|0[1-9][0-9]|00[1-9]', 'WW':
    '5[0-2]|[1-4][0-9]|[0-9]', '0W': '5[0-2]|[0-4][0-9]', 'UU':
    '5[0-2]|[1-4][0-9]|[0-9]', '0U': '5[0-2]|[0-4][0-9]', 'VV':
    '5[0-3]|[1-4][0-9]|[1-9]', '0V': '5[0-3]|[1-4][0-9]|0[1-9]', 'MAJOR':
    '[0-9]+', 'MINOR': '[0-9]+', 'PATCH': '[0-9]+', 'BUILD': '[0-9]+',
    'BLD': '[1-9][0-9]*', 'TAG': 'preview|final|alpha|beta|post|rc',
    'PYTAG': 'post|rc|a|b', 'NUM': '[0-9]+', 'INC0': '[0-9]+', 'INC1':
    '[1-9][0-9]*'}
PATTERN_PART_FIELDS = {'YYYY': 'year_y', 'YY': 'year_y', '0Y': 'year_y',
    'GGGG': 'year_g', 'GG': 'year_g', '0G': 'year_g', 'Q': 'quarter', 'MM':
    'month', '0M': 'month', 'DD': 'dom', '0D': 'dom', 'JJJ': 'doy', '00J':
    'doy', 'MAJOR': 'major', 'MINOR': 'minor', 'PATCH': 'patch', 'BUILD':
    'bid', 'BLD': 'bid', 'TAG': 'tag', 'PYTAG': 'pytag', 'NUM': 'num',
    'INC0': 'inc0', 'INC1': 'inc1', 'WW': 'week_w', '0W': 'week_w', 'UU':
    'week_u', '0U': 'week_u', 'VV': 'week_v', '0V': 'week_v'}
PEP440_PART_SUBSTITUTIONS = {'0W': 'WW', '0U': 'UU', '0V': 'VV', '0M': 'MM',
    '0D': 'DD', '00J': 'JJJ', 'BUILD': 'BLD', 'TAG': 'PYTAG'}
FieldValue = typ.Union[str, int]


def _fmt_num(val):
    return str(val)


def _fmt_bld(val):
    return str(int(val))


def _fmt_yy(year_y):
    return str(int(str(year_y)[-2:]))


def _fmt_0y(year_y):
    return '{0:02}'.format(int(str(year_y)[-2:]))


def _fmt_gg(year_g):
    return str(int(str(year_g)[-2:]))


def _fmt_0g(year_g):
    return '{0:02}'.format(int(str(year_g)[-2:]))


def _fmt_0m(month):
    return '{0:02}'.format(int(month))


def _fmt_0d(dom):
    return '{0:02}'.format(int(dom))


def _fmt_00j(doy):
    return '{0:03}'.format(int(doy))


def _fmt_0w(week_w):
    return '{0:02}'.format(int(week_w))


def _fmt_0u(week_u):
    return '{0:02}'.format(int(week_u))


def _fmt_0v(week_v):
    return '{0:02}'.format(int(week_v))


FormatterFunc = typ.Callable[[FieldValue], str]
PART_FORMATS = {'YYYY': _fmt_num, 'YY': _fmt_yy, '0Y': _fmt_0y, 'GGGG':
    _fmt_num, 'GG': _fmt_gg, '0G': _fmt_0g, 'Q': _fmt_num, 'MM': _fmt_num,
    '0M': _fmt_0m, 'DD': _fmt_num, '0D': _fmt_0d, 'JJJ': _fmt_num, '00J':
    _fmt_00j, 'MAJOR': _fmt_num, 'MINOR': _fmt_num, 'PATCH': _fmt_num,
    'BUILD': _fmt_num, 'BLD': _fmt_bld, 'TAG': _fmt_num, 'PYTAG': _fmt_num,
    'NUM': _fmt_num, 'INC0': _fmt_num, 'INC1': _fmt_num, 'WW': _fmt_num,
    '0W': _fmt_0w, 'UU': _fmt_num, '0U': _fmt_0u, 'VV': _fmt_num, '0V': _fmt_0v
    }


def _convert_to_pep440(version_pattern):
    pep440_pattern = version_pattern
    if pep440_pattern.startswith('v'):
        pep440_pattern = pep440_pattern[1:]
    pep440_pattern = pep440_pattern.replace('\\[', '')
    pep440_pattern = pep440_pattern.replace('\\]', '')
    pep440_pattern, _ = re.subn('[^a-zA-Z0-9\\.\\[\\]]', '', pep440_pattern)
    part_names = list(PATTERN_PART_FIELDS.keys())
    part_names.sort(key=len, reverse=True)
    for part_name in part_names:
        if part_name not in version_pattern:
            continue
        if part_name not in PEP440_PART_SUBSTITUTIONS:
            continue
        substitution = PEP440_PART_SUBSTITUTIONS[part_name]
        if substitution in pep440_pattern:
            continue
        is_numerical_part = part_name not in ('TAG', 'PYTAG')
        if is_numerical_part:
            part_index = pep440_pattern.find(part_name)
            is_zero_truncation_part = part_index == 0 or pep440_pattern[
                part_index - 1] == '.'
            if is_zero_truncation_part:
                pep440_pattern = pep440_pattern.replace(part_name, substitution
                    )
        else:
            pep440_pattern = pep440_pattern.replace(part_name, substitution)
    if 'PYTAGNUM' not in pep440_pattern:
        pep440_pattern = pep440_pattern.replace('PYTAG', '')
        pep440_pattern = pep440_pattern.replace('NUM', '')
        pep440_pattern = pep440_pattern.replace('[]', '')
        pep440_pattern += '[PYTAGNUM]'
    return pep440_pattern


def normalize_pattern(version_pattern, raw_pattern):
    normalized_pattern = raw_pattern
    if '{version}' in raw_pattern:
        normalized_pattern = normalized_pattern.replace('{version}',
            version_pattern)
    if '{pep440_version}' in normalized_pattern:
        pep440_version_pattern = _convert_to_pep440(version_pattern)
        normalized_pattern = normalized_pattern.replace('{pep440_version}',
            pep440_version_pattern)
    return normalized_pattern


def _replace_pattern_parts(pattern):
    while True:
        new_pattern, _n = re.subn('([^\\\\]|^)\\[', '\\1(?:', pattern)
        new_pattern, _m = re.subn('([^\\\\]|^)\\]', '\\1)?', new_pattern)
        pattern = new_pattern
        if _n + _m == 0:
            break
    SortKey = typ.Tuple[int, int]
    PostitionedPart = typ.Tuple[int, int, str]
    part_patterns_by_index = {}
    for part_name, part_pattern in PART_PATTERNS.items():
        start_idx = pattern.find(part_name)
        if start_idx >= 0:
            field = PATTERN_PART_FIELDS[part_name]
            named_part_pattern = '(?P<{0}>{1})'.format(field, part_pattern)
            end_idx = start_idx + len(part_name)
            sort_key = -end_idx, -len(part_name)
            part_patterns_by_index[sort_key
                ] = start_idx, end_idx, named_part_pattern
    last_start_idx = len(pattern) + 1
    result_pattern = pattern
    for _, (start_idx, end_idx, named_part_pattern) in sorted(
        part_patterns_by_index.items()):
        if end_idx <= last_start_idx:
            result_pattern = result_pattern[:start_idx
                ] + named_part_pattern + result_pattern[end_idx:]
            last_start_idx = start_idx
    return result_pattern


def _compile_pattern_re(normalized_pattern):
    escaped_pattern = normalized_pattern
    for char, escaped in RE_PATTERN_ESCAPES:
        is_semantic_char = char in '[]\\'
        if not is_semantic_char:
            escaped_pattern = escaped_pattern.replace(char, escaped)
    pattern_str = _replace_pattern_parts(escaped_pattern)
    return re.compile(pattern_str)


@utils.memo
def compile_pattern(version_pattern, raw_pattern=None):
    _raw_pattern = version_pattern if raw_pattern is None else raw_pattern
    normalized_pattern = normalize_pattern(version_pattern, _raw_pattern)
    regexp = _compile_pattern_re(normalized_pattern)
    return Pattern(version_pattern, normalized_pattern, regexp)


def compile_patterns(version_pattern, raw_patterns):
    return [compile_pattern(version_pattern, raw_pattern) for raw_pattern in
        raw_patterns]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the lexid project
# https://github.com/mbarkhau/lexid
#
# Copyright (c) 2020-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import sys
import click
from lexid import next_id
try:
    import builtins
except ImportError:
    import __builtin__ as builtins
from lexid import ord_val
range = getattr(builtins, 'xrange', range)
try:
    import pretty_traceback
    pretty_traceback.install()
except ImportError:
    pass
click.disable_unicode_literals_warning = True


@click.command()
@click.option('-n', '--num', default=1)
@click.option('--debug', is_flag=True, default=False)
@click.argument('start_id', default='1001')
def main(start_id='1001', num=1, debug=False):
    """Increment a lexid."""
    if debug:
        print('{0:<13} {1:>12}'.format('lexical', 'numerical'))
    _curr_id = start_id
    for _ in range(num):
        try:
            _next_id = next_id(_curr_id)
        except OverflowError as err:
            sys.stderr.write('OverflowError: {0}'.format(err))
            sys.exit(1)
        if debug:
            print('{0:<13} {1:>12}'.format(_next_id, ord_val(_next_id)))
        else:
            print(_next_id)
        _curr_id = _next_id



# -*- coding: utf-8 -*-
# This file is part of the bumpver project
# https://github.com/mbarkhau/bumpver
#
# Copyright (c) 2018-2022 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
"""Parse PyCalVer strings from files."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import typing as typ
try:
    import builtins
except ImportError:
    import __builtin__ as builtins
from .patterns import Pattern
str = getattr(builtins, 'unicode', str)
LineNo = int
Start = int
End = int
LineSpan = typ.NamedTuple('LineSpan', [('lineno', LineNo), ('start', Start),
    ('end', End)])
LineSpans = typ.List[LineSpan]


def _has_overlap(needle, haystack):
    for span in haystack:
        has_overlap = (span.lineno == needle.lineno and needle.start <=
            span.end and needle.end >= span.start)
        if has_overlap:
            return True
    return False


PatternMatch = typ.NamedTuple('PatternMatch', [('lineno', LineNo), ('line',
    str), ('pattern', Pattern), ('span', typ.Tuple[Start, End]), ('match',
    str)])
PatternMatches = typ.Iterable[PatternMatch]


def _iter_for_pattern(lines, pattern):
    for lineno, line in enumerate(lines):
        match = pattern.regexp.search(line)
        if match and len(match.group(0)) > 0:
            yield PatternMatch(lineno, line, pattern, match.span(), match.
                group(0))


def iter_matches(lines, patterns):
    """Iterate over all matches of any pattern on any line.

    >>> from . import v1patterns
    >>> lines = ["__version__ = 'v201712.0002-alpha'"]
    >>> version_pattern = "{pycalver}"
    >>> raw_patterns = ["{pycalver}", "{pep440_pycalver}"]
    >>> patterns = [v1patterns.compile_pattern(version_pattern, p) for p in raw_patterns]
    >>> matches = list(iter_matches(lines, patterns))
    >>> assert matches[0] == PatternMatch(
    ...     lineno = 0,
    ...     line   = "__version__ = 'v201712.0002-alpha'",
    ...     pattern= v1patterns.compile_pattern(version_pattern),
    ...     span   = (15, 33),
    ...     match  = "v201712.0002-alpha",
    ... )
    """
    matched_spans = []
    for pattern in patterns:
        for match in _iter_for_pattern(lines, pattern):
            needle_span = LineSpan(match.lineno, *match.span)
            if not _has_overlap(needle_span, matched_spans):
                yield match
            matched_spans.append(needle_span)



# -*- coding: utf-8 -*-
# This file is part of the bumpver project
# https://github.com/mbarkhau/bumpver
#
# Copyright (c) 2018-2022 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import typing as typ
import difflib
try:
    import builtins
except ImportError:
    import __builtin__ as builtins
import pathlib2 as pl
str = getattr(builtins, 'unicode', str)
from . import config
from .patterns import Pattern


class NoPatternMatch(Exception):
    """Pattern not found in content.

    logger.error is used to show error info about the patterns so
    that users can debug what is wrong with them. The class
    itself doesn't capture that info. This approach is used so
    that all patter issues can be shown, rather than bubbling
    all the way up the stack on the very first pattern with no
    matches.
    """


def detect_line_sep(content):
    """Parse line separator from content.

    >>> detect_line_sep('\\r\\n')
    '\\r\\n'
    >>> detect_line_sep('\\r')
    '\\r'
    >>> detect_line_sep('\\n')
    '\\n'
    >>> detect_line_sep('')
    '\\n'
    """
    if '\r\n' in content:
        return '\r\n'
    elif '\r' in content:
        return '\r'
    else:
        return '\n'


RewrittenFileData = typ.NamedTuple('RewrittenFileData', [('path', str), (
    'line_sep', str), ('old_lines', typ.List[str]), ('new_lines', typ.List[
    str])])
PathPatternsItem = typ.Tuple[pl.Path, typ.List[Pattern]]


def iter_path_patterns_items(file_patterns):
    for filepath_str, patterns in file_patterns.items():
        filepath_obj = pl.Path(filepath_str)
        if filepath_obj.exists():
            yield filepath_obj, patterns
        else:
            errmsg = "File does not exist: '{0}'".format(filepath_str)
            raise IOError(errmsg)


def diff_lines(rfd):
    """Generate unified diff.

    >>> rfd = RewrittenFileData(
    ...    path      = "<path>",
    ...    line_sep  = "\\n",
    ...    old_lines = ["foo"],
    ...    new_lines = ["bar"],
    ... )
    >>> diff_lines(rfd)
    ['--- <path>', '+++ <path>', '@@ -1 +1 @@', '-foo', '+bar']
    """
    lines = difflib.unified_diff(a=rfd.old_lines, b=rfd.new_lines, lineterm
        ='', fromfile=rfd.path, tofile=rfd.path)
    return list(lines)



# -*- coding: utf-8 -*-
# This file is part of the bumpver project
# https://github.com/mbarkhau/bumpver
#
# Copyright (c) 2018-2022 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
try:
    import builtins
except ImportError:
    import __builtin__ as builtins
import typing as typ
str = getattr(builtins, 'unicode', str)
Pattern = typ.NamedTuple('Pattern', [('version_pattern', str), (
    'raw_pattern', str), ('regexp', typ.Pattern[str])])
RE_PATTERN_ESCAPES = [('\\', '\\\\'), ('-', '\\-'), ('.', '\\.'), ('+',
    '\\+'), ('*', '\\*'), ('?', '\\?'), ('{', '\\{'), ('}', '\\}'), ('[',
    '\\['), (']', '\\]'), ('(', '\\('), (')', '\\)')]



# -*- coding: utf-8 -*-
# This file is part of the bumpver project
# https://github.com/mbarkhau/bumpver
#
# Copyright (c) 2018-2022 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
"""Functions related to version string manipulation."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import typing as typ
import logging
import datetime as dt
try:
    import builtins
except ImportError:
    import __builtin__ as builtins
import lexid
str = getattr(builtins, 'unicode', str)
from . import version
from . import v1patterns
logger = logging.getLogger('bumpver.v1version')
CalInfo = typ.Union[version.V1CalendarInfo, version.V1VersionInfo]


def _is_cal_gt(left, right):
    """Is left > right for non-None fields."""
    lvals = []
    rvals = []
    for field in version.V1CalendarInfo._fields:
        lval = getattr(left, field)
        rval = getattr(right, field)
        if not (lval is None or rval is None):
            lvals.append(lval)
            rvals.append(rval)
    return lvals > rvals


def _ver_to_cal_info(vnfo):
    return version.V1CalendarInfo(vnfo.year, vnfo.quarter, vnfo.month, vnfo
        .dom, vnfo.doy, vnfo.iso_week, vnfo.us_week)


def cal_info(date=None):
    """Generate calendar components for current date.

    >>> from datetime import date

    >>> c = cal_info(date(2019, 1, 5))
    >>> (c.year, c.quarter, c.month, c.dom, c.doy, c.iso_week, c.us_week)
    (2019, 1, 1, 5, 5, 0, 0)

    >>> c = cal_info(date(2019, 1, 6))
    >>> (c.year, c.quarter, c.month, c.dom, c.doy, c.iso_week, c.us_week)
    (2019, 1, 1, 6, 6, 0, 1)

    >>> c = cal_info(date(2019, 1, 7))
    >>> (c.year, c.quarter, c.month, c.dom, c.doy, c.iso_week, c.us_week)
    (2019, 1, 1, 7, 7, 1, 1)

    >>> c = cal_info(date(2019, 4, 7))
    >>> (c.year, c.quarter, c.month, c.dom, c.doy, c.iso_week, c.us_week)
    (2019, 2, 4, 7, 97, 13, 14)
    """
    if date is None:
        date = version.TODAY
    kwargs = {'year': date.year, 'quarter': version.quarter_from_month(date
        .month), 'month': date.month, 'dom': date.day, 'doy': int(date.
        strftime('%j'), base=10), 'iso_week': int(date.strftime('%W'), base
        =10), 'us_week': int(date.strftime('%U'), base=10)}
    return version.V1CalendarInfo(**kwargs)


FieldKey = str
MatchGroupKey = str
MatchGroupStr = str
PatternGroups = typ.Dict[MatchGroupKey, MatchGroupStr]
FieldValues = typ.Dict[FieldKey, MatchGroupStr]


def _parse_field_values(field_values):
    fvals = field_values
    tag = fvals.get('tag')
    if tag is None:
        tag = 'final'
    tag = version.TAG_BY_PEP440_TAG.get(tag, tag)
    assert tag is not None
    bid = fvals['bid'] if 'bid' in fvals else '0001'
    year = int(fvals['year']) if 'year' in fvals else None
    if year is not None and year < 100:
        year += 2000
    doy = int(fvals['doy']) if 'doy' in fvals else None
    month = None
    dom = None
    if year and doy:
        date = version.date_from_doy(year, doy)
        month = date.month
        dom = date.day
    else:
        month = int(fvals['month']) if 'month' in fvals else None
        dom = int(fvals['dom']) if 'dom' in fvals else None
    iso_week = None
    us_week = None
    if year and month and dom:
        date = dt.date(year, month, dom)
        doy = int(date.strftime('%j'), base=10)
        iso_week = int(date.strftime('%W'), base=10)
        us_week = int(date.strftime('%U'), base=10)
    else:
        iso_week = None
        us_week = None
    quarter = int(fvals['quarter']) if 'quarter' in fvals else None
    if quarter is None and month:
        quarter = version.quarter_from_month(month)
    major = int(fvals['major']) if 'major' in fvals else 0
    minor = int(fvals['minor']) if 'minor' in fvals else 0
    patch = int(fvals['patch']) if 'patch' in fvals else 0
    return version.V1VersionInfo(year=year, quarter=quarter, month=month,
        dom=dom, doy=doy, iso_week=iso_week, us_week=us_week, major=major,
        minor=minor, patch=patch, bid=bid, tag=tag)


def _is_calver(cinfo):
    """Check pattern for any calendar based parts.

    >>> _is_calver(cal_info())
    True

    >>> vnfo = _parse_version_info({'year': "2018", 'month': "11", 'bid': "0018"})
    >>> _is_calver(vnfo)
    True

    >>> vnfo = _parse_version_info({'MAJOR': "1", 'MINOR': "023", 'PATCH': "45"})
    >>> _is_calver(vnfo)
    False
    """
    for field in version.V1CalendarInfo._fields:
        maybe_val = getattr(cinfo, field, None)
        if isinstance(maybe_val, int):
            return True
    return False


VersionInfoKW = typ.Dict[str, typ.Union[str, int, None]]


def _parse_pattern_groups(pattern_groups):
    for part_name in pattern_groups.keys():
        is_valid_part_name = (part_name in v1patterns.
            COMPOSITE_PART_PATTERNS or part_name in v1patterns.
            PATTERN_PART_FIELDS)
        if not is_valid_part_name:
            err_msg = "Invalid part '{0}'".format(part_name)
            raise version.PatternError(err_msg)
    field_value_items = [(field_name, pattern_groups[part_name]) for 
        part_name, field_name in v1patterns.PATTERN_PART_FIELDS.items() if 
        part_name in pattern_groups.keys()]
    all_fields = [field_name for field_name, _ in field_value_items]
    unique_fields = set(all_fields)
    duplicate_fields = [f for f in unique_fields if all_fields.count(f) > 1]
    if any(duplicate_fields):
        err_msg = 'Multiple parts for same field {0}.'.format(duplicate_fields)
        raise version.PatternError(err_msg)
    else:
        return dict(field_value_items)


def _parse_version_info(pattern_groups):
    """Parse normalized V1VersionInfo from groups of a matched pattern.

    >>> vnfo = _parse_version_info({'year': "2018", 'month': "11", 'bid': "0099"})
    >>> (vnfo.year, vnfo.month, vnfo.quarter, vnfo.bid, vnfo.tag)
    (2018, 11, 4, '0099', 'final')

    >>> vnfo = _parse_version_info({'year': "18", 'month': "11"})
    >>> (vnfo.year, vnfo.month, vnfo.quarter)
    (2018, 11, 4)

    >>> vnfo = _parse_version_info({'year': "2018", 'doy': "11", 'bid': "099", 'tag': "b"})
    >>> (vnfo.year, vnfo.month, vnfo.dom, vnfo.bid, vnfo.tag)
    (2018, 1, 11, '099', 'beta')

    >>> vnfo = _parse_version_info({'MAJOR': "1", 'MINOR': "23", 'PATCH': "45"})
    >>> (vnfo.major, vnfo.minor, vnfo.patch)
    (1, 23, 45)

    >>> vnfo = _parse_version_info({'MAJOR': "1", 'MMM': "023", 'PPPP': "0045"})
    >>> (vnfo.major, vnfo.minor, vnfo.patch)
    (1, 23, 45)
    """
    field_values = _parse_pattern_groups(pattern_groups)
    return _parse_field_values(field_values)


def parse_version_info(version_str, raw_pattern='{pycalver}'):
    """Parse normalized V1VersionInfo.

    >>> vnfo = parse_version_info("v201712.0033-beta", raw_pattern="{pycalver}")
    >>> assert vnfo == _parse_version_info({'year': 2017, 'month': 12, 'bid': "0033", 'tag': "beta"})

    >>> vnfo = parse_version_info("1.23.456", raw_pattern="{semver}")
    >>> assert vnfo == _parse_version_info({'MAJOR': "1", 'MINOR': "23", 'PATCH': "456"})
    """
    pattern = v1patterns.compile_pattern(raw_pattern)
    match = pattern.regexp.match(version_str)
    if match is None:
        err_msg = ("Invalid version string '{0}' for pattern '{1}'/'{2}'".
            format(version_str, raw_pattern, pattern.regexp.pattern))
        raise version.PatternError(err_msg)
    else:
        return _parse_version_info(match.groupdict())


def is_valid(version_str, raw_pattern='{pycalver}'):
    """Check if a version matches a pattern.

    >>> is_valid("v201712.0033-beta", raw_pattern="{pycalver}")
    True
    >>> is_valid("v201712.0033-beta", raw_pattern="{semver}")
    False
    >>> is_valid("1.2.3", raw_pattern="{semver}")
    True
    >>> is_valid("v201712.0033-beta", raw_pattern="{semver}")
    False
    """
    try:
        parse_version_info(version_str, raw_pattern)
        return True
    except version.PatternError:
        return False


ID_FIELDS_BY_PART = {'MAJOR': 'major', 'MINOR': 'minor', 'MM': 'minor',
    'MMM': 'minor', 'MMMM': 'minor', 'MMMMM': 'minor', 'MMMMMM': 'minor',
    'MMMMMMM': 'minor', 'PATCH': 'patch', 'PP': 'patch', 'PPP': 'patch',
    'PPPP': 'patch', 'PPPPP': 'patch', 'PPPPPP': 'patch', 'PPPPPPP':
    'patch', 'BID': 'bid', 'BB': 'bid', 'BBB': 'bid', 'BBBB': 'bid',
    'BBBBB': 'bid', 'BBBBBB': 'bid', 'BBBBBBB': 'bid'}


def format_version(vinfo, raw_pattern):
    """Generate version string.

    >>> import datetime as dt
    >>> vinfo = parse_version_info("v201712.0033-beta", raw_pattern="{pycalver}")
    >>> vinfo_a = vinfo._replace(**cal_info(date=dt.date(2017, 1, 1))._asdict())
    >>> vinfo_b = vinfo._replace(**cal_info(date=dt.date(2017, 12, 31))._asdict())
    >>> vinfo_c = vinfo_b._replace(major=1, minor=2, patch=34, tag='final')

    >>> format_version(vinfo_a, raw_pattern="v{yy}.{BID}{release}")
    'v17.33-beta'
    >>> format_version(vinfo_a, raw_pattern="{pep440_version}")
    '201701.33b0'

    >>> format_version(vinfo_a, raw_pattern="{pycalver}")
    'v201701.0033-beta'
    >>> format_version(vinfo_b, raw_pattern="{pycalver}")
    'v201712.0033-beta'

    >>> format_version(vinfo_a, raw_pattern="v{year}w{iso_week}.{BID}{release}")
    'v2017w00.33-beta'
    >>> format_version(vinfo_b, raw_pattern="v{year}w{iso_week}.{BID}{release}")
    'v2017w52.33-beta'

    >>> format_version(vinfo_a, raw_pattern="v{year}d{doy}.{bid}{release}")
    'v2017d001.0033-beta'
    >>> format_version(vinfo_b, raw_pattern="v{year}d{doy}.{bid}{release}")
    'v2017d365.0033-beta'

    >>> format_version(vinfo_c, raw_pattern="v{year}w{iso_week}.{BID}-{tag}")
    'v2017w52.33-final'
    >>> format_version(vinfo_c, raw_pattern="v{year}w{iso_week}.{BID}{release}")
    'v2017w52.33'

    >>> format_version(vinfo_c, raw_pattern="v{MAJOR}.{MINOR}.{PATCH}")
    'v1.2.34'
    >>> format_version(vinfo_c, raw_pattern="v{MAJOR}.{MM}.{PPP}")
    'v1.02.034'
    """
    full_pattern = raw_pattern
    for part_name, full_part_format in v1patterns.FULL_PART_FORMATS.items():
        full_pattern = full_pattern.replace('{' + part_name + '}',
            full_part_format)
    kwargs = vinfo._asdict()
    release_tag = vinfo.tag
    if release_tag == 'final':
        kwargs['release'] = ''
        kwargs['pep440_tag'] = ''
    else:
        kwargs['release'] = '-' + release_tag
        kwargs['pep440_tag'] = version.PEP440_TAG_BY_TAG[release_tag] + '0'
    kwargs['release_tag'] = release_tag
    year = vinfo.year
    if year:
        kwargs['yy'] = str(year)[-2:]
        kwargs['yyyy'] = year
    kwargs['BID'] = int(vinfo.bid, 10)
    for part_name, field in ID_FIELDS_BY_PART.items():
        val = kwargs[field]
        if part_name.lower() == field.lower():
            if isinstance(val, str):
                kwargs[part_name] = int(val, base=10)
            else:
                kwargs[part_name] = val
        else:
            assert len(set(part_name)) == 1
            padded_len = len(part_name)
            kwargs[part_name] = str(val).zfill(padded_len)
    return full_pattern.format(**kwargs)


def incr(old_version, raw_pattern='{pycalver}', **kwargs):
    major = kwargs.get('major', False)
    minor = kwargs.get('minor', False)
    patch = kwargs.get('patch', False)
    tag = kwargs.get('tag', None)
    tag_num = kwargs.get('tag_num', False)
    pin_date = kwargs.get('pin_date', False)
    maybe_date = kwargs.get('maybe_date', None)
    """Increment version string.

    'old_version' is assumed to be a string that matches 'pattern'
    """
    try:
        old_vinfo = parse_version_info(old_version, raw_pattern)
    except version.PatternError as ex:
        logger.error(str(ex))
        return None
    date = version.TODAY if maybe_date is None else maybe_date
    cur_cinfo = _ver_to_cal_info(old_vinfo) if pin_date else cal_info(date)
    if _is_cal_gt(old_vinfo, cur_cinfo):
        logger.warning("Old version appears to be from the future '{0}'".
            format(old_version))
        cur_vinfo = old_vinfo
    else:
        cur_vinfo = old_vinfo._replace(**cur_cinfo._asdict())
    cur_vinfo = cur_vinfo._replace(bid=lexid.next_id(cur_vinfo.bid))
    if major:
        cur_vinfo = cur_vinfo._replace(major=cur_vinfo.major + 1, minor=0,
            patch=0)
    if minor:
        cur_vinfo = cur_vinfo._replace(minor=cur_vinfo.minor + 1, patch=0)
    if patch:
        cur_vinfo = cur_vinfo._replace(patch=cur_vinfo.patch + 1)
    if tag_num:
        raise NotImplementedError(
            '--tag-num not supported for old style patterns')
    if tag:
        cur_vinfo = cur_vinfo._replace(tag=tag)
    new_version = format_version(cur_vinfo, raw_pattern)
    if new_version == old_version:
        logger.error('Invalid arguments or pattern, version did not change.')
        return None
    else:
        return new_version



def get_object_by_keys(dictionary, keys):
  if not keys: return None
  if type(keys)==list:
    key = keys.pop(0)
    result = dictionary.get(key)
    if not result: return result
    if not keys: return result
    return get_object_by_keys(result, keys)
  else:
    return dictionary.get(keys)



# -*- coding: utf-8 -*-
# This file is part of the bumpver project
# https://github.com/mbarkhau/bumpver
#
# Copyright (c) 2018-2022 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import re
import logging
import textwrap
from . import pysix
logger = logging.getLogger('bumpver.regexfmt')


def format_regex(regex):
    """Format a regex pattern suitible for flags=re.VERBOSE.

    >>> regex = r"\\[CalVer v(?P<year_y>[1-9][0-9]{3})(?P<month>(?:1[0-2]|0[1-9]))"
    >>> print(format_regex(regex))
    \\[CalVer[ ]v
    (?P<year_y>[1-9][0-9]{3})
    (?P<month>
        (?:1[0-2]|0[1-9])
    )
    """
    re.compile(regex)
    tmp_regex = regex.replace(' ', '[ ]')
    tmp_regex = tmp_regex.replace('"', '\\"')
    tmp_regex, _ = re.subn('([^\\\\])?\\)(\\?)?', '\\1)\\2\n', tmp_regex)
    tmp_regex, _ = re.subn('([^\\\\])\\(', '\\1\n(', tmp_regex)
    tmp_regex, _ = re.subn('^\\)\\)', ')\n)', tmp_regex, flags=re.MULTILINE)
    lines = tmp_regex.splitlines()
    indented_lines = []
    level = 0
    for line in lines:
        if line.strip():
            increment = line.count('(') - line.count(')')
            if increment >= 0:
                line = '    ' * level + line
                level += increment
            else:
                level += increment
                line = '    ' * level + line
            indented_lines.append(line)
    formatted_regex = '\n'.join(indented_lines)
    re.compile(formatted_regex)
    return formatted_regex


def pyexpr_regex(regex):
    try:
        formatted_regex = format_regex(regex)
        formatted_regex = textwrap.indent(formatted_regex.rstrip(), '    ')
        return ('re.compile(r"""\n' + formatted_regex +
            '\n""", flags=re.VERBOSE)')
    except re.error:
        return 're.compile({0})'.format(repr(regex))


URL_TEMPLATE = 'https://regex101.com/?flavor=python&flags=gmx&regex='


def regex101_url(regex_pattern):
    try:
        regex_pattern = format_regex(regex_pattern)
    except re.error:
        logger.warning("Error formatting regex '{0}'".format(repr(
            regex_pattern)))
    return URL_TEMPLATE + pysix.quote(regex_pattern)



from setuptools import setup, find_packages
import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

classifier = [
  'Development Status :: Production',
  'Intended Audience :: Developers',
  'License :: OSI Approved :: MIT License',
  'programming Language :: Python :: 3.8'
]

def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ) as fh:
        return fh.read()


setup(
  name='communicationAppModule',
  version="0.2.2",
  description='Module to connect with communication app and send message',
  long_description="%s\n%s" % (
        re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', read('README.md')),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.txt'))
    ),
  long_description_content_type="text/markdown",
  author='Wayne Shang',
  author_email='wayne18308@gmail.com',
  license='MIT',
  classifier=classifier,
  keywords='Communication App',
  packages=find_packages(),
  install_requires=['']
)


from communicationApp import createClient
def main():

  telegram = createClient('telegram', bot_token ='1632387033:AAEgUxJiAwZBRVXwVpocXohQOxZYFhkkR6g')
  a = {'image_url': "https://im.uniqlo.com/style/SDP1779_30012719.jpg"}
  result = telegram.send_photo(image_url="https://im.uniqlo.com/style/SDP1779_30012719.jpg")
  print("---------end")
  print(result)

if __name__ == '__main__':
  try:
      main()
  except KeyboardInterrupt:
      exit()


# -*- coding: utf-8 -*-
# This file is part of the bumpver project
# https://github.com/mbarkhau/bumpver
#
# Copyright (c) 2018-2022 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import typing as typ
try:
    import builtins
except ImportError:
    import __builtin__ as builtins
import datetime as dt
str = getattr(builtins, 'unicode', str)
MaybeInt = typ.Optional[int]


def parse_version(version):
    import pkg_resources
    return pkg_resources.parse_version(version)


V1CalendarInfo = typ.NamedTuple('V1CalendarInfo', [('year', MaybeInt), (
    'quarter', MaybeInt), ('month', MaybeInt), ('dom', MaybeInt), ('doy',
    MaybeInt), ('iso_week', MaybeInt), ('us_week', MaybeInt)])
V1VersionInfo = typ.NamedTuple('V1VersionInfo', [('year', MaybeInt), (
    'quarter', MaybeInt), ('month', MaybeInt), ('dom', MaybeInt), ('doy',
    MaybeInt), ('iso_week', MaybeInt), ('us_week', MaybeInt), ('major', int
    ), ('minor', int), ('patch', int), ('bid', str), ('tag', str)])
V2CalendarInfo = typ.NamedTuple('V2CalendarInfo', [('year_y', MaybeInt), (
    'year_g', MaybeInt), ('quarter', MaybeInt), ('month', MaybeInt), ('dom',
    MaybeInt), ('doy', MaybeInt), ('week_w', MaybeInt), ('week_u', MaybeInt
    ), ('week_v', MaybeInt)])
V2VersionInfo = typ.NamedTuple('V2VersionInfo', [('year_y', MaybeInt), (
    'year_g', MaybeInt), ('quarter', MaybeInt), ('month', MaybeInt), ('dom',
    MaybeInt), ('doy', MaybeInt), ('week_w', MaybeInt), ('week_u', MaybeInt
    ), ('week_v', MaybeInt), ('major', int), ('minor', int), ('patch', int),
    ('bid', str), ('tag', str), ('pytag', str), ('num', int), ('inc0', int),
    ('inc1', int)])
TODAY = dt.datetime.utcnow().date()
TAG_BY_PEP440_TAG = {'a': 'alpha', 'b': 'beta', '': 'final', 'rc': 'rc',
    'dev': 'dev', 'post': 'post'}
PEP440_TAG_BY_TAG = {'a': 'a', 'b': 'b', 'dev': 'dev', 'alpha': 'a', 'beta':
    'b', 'preview': 'rc', 'pre': 'rc', 'rc': 'rc', 'c': 'rc', 'final': '',
    'post': 'post', 'r': 'post', 'rev': 'post'}
assert set(TAG_BY_PEP440_TAG.keys()) == set(PEP440_TAG_BY_TAG.values())
assert set(TAG_BY_PEP440_TAG.values()) < set(PEP440_TAG_BY_TAG.keys())
PART_ZERO_VALUES = {'MAJOR': '0', 'MINOR': '0', 'PATCH': '0', 'TAG':
    'final', 'PYTAG': '', 'NUM': '0', 'INC0': '0'}
V2_FIELD_INITIAL_VALUES = {'major': '0', 'minor': '0', 'patch': '0', 'num':
    '0', 'inc0': '0', 'inc1': '1'}


def is_zero_val(part, part_value):
    return part in PART_ZERO_VALUES and part_value == PART_ZERO_VALUES[part]


class PatternError(Exception):
    pass


def date_from_doy(year, doy):
    """Parse date from year and day of year (1 indexed).

    >>> cases = [
    ...     (2016, 1), (2016, 31), (2016, 31 + 1), (2016, 31 + 29), (2016, 31 + 30),
    ...     (2017, 1), (2017, 31), (2017, 31 + 1), (2017, 31 + 28), (2017, 31 + 29),
    ... ]
    >>> dates = [date_from_doy(year, month) for year, month in cases]
    >>> assert [(d.month, d.day) for d in dates] == [
    ...     (1, 1), (1, 31), (2, 1), (2, 29), (3, 1),
    ...     (1, 1), (1, 31), (2, 1), (2, 28), (3, 1),
    ... ]
    """
    return dt.date(year, 1, 1) + dt.timedelta(days=doy - 1)


def quarter_from_month(month):
    """Calculate quarter (1 indexed) from month (1 indexed).

    >>> [quarter_from_month(month) for month in range(1, 13)]
    [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4]
    """
    return (month - 1) // 3 + 1


def to_pep440(version):
    """Derive pep440 compliant version string from PyCalVer version string.

    >>> to_pep440("v201811.0007-beta")
    '201811.7b0'
    """
    return str(parse_version(version))




import requests
import datetime
import logging
import sqlite3
import asyncio
from communicationApp.utils import get_object_by_keys 
from communicationApp.communicationApp import CommunicationApp
import os

TELEGRAM_API = "https://api.telegram.org/bot"
GET_UPDATE = 'getUpdates'
SEND_MESSAGE = 'sendMessage'
SEND_PHOTO= 'sendPhoto'
DEFAULT_DB = 'db/telegram.db'
DEFAULT_LOG_FILE = f"log/{datetime.date.today().strftime('%Y-%m-%d')}.log"
# 400 user error
# 404 exception
class Telegram(CommunicationApp):
  def __init__(self, **kwargs):
    bot_token = get_object_by_keys(kwargs, 'bot_token')
    db_name = get_object_by_keys(kwargs, 'db_name') or DEFAULT_DB
    log_file_name = get_object_by_keys(kwargs, 'log_file_name') or DEFAULT_LOG_FILE

    super().__init__(type(self).__name__)
    self.bot_token = bot_token
    self.bot_url = f"{TELEGRAM_API}{bot_token}/"
    self.db_name = db_name
    self.log_file_name = log_file_name
    self._create_folder(os.path.dirname(log_file_name))
    # db
    self._create_folder(os.path.dirname(db_name))
    self.connection = self.init_db()
    self.cursor = self.connection.cursor()
    self._create_table()
    self.chat_id = self._get_chat_id()
    

  def _create_folder(self, directory_name):
    try:
      os.stat(directory_name)
    except:
      os.mkdir(directory_name)

  def init_db(self):
    connection = sqlite3.connect(self.db_name)
    return connection
  
  def send_message(self, text):
    try:
      # print("chat_id----------", self.chat_id)
      result = {}
      if self.chat_id is None: return {"status_code": 400, "error": "Please send a message to bot"}
      if type(text) == list:
        result["status_code"] = 200
        result["data"] = []
        for t in text:
          message_result = self.send_message(t)
          if message_result["status_code"] == 200: 
            result["data"] = result["data"] + message_result["data"]
          else:
            return {**result, **message_result}
      else:
        
        params = {'chat_id': self.chat_id, 'text': text, 'parse_mode': 'HTML'}
        method = SEND_MESSAGE
        response = requests.post(self.bot_url + method, params)
        result = self._handle_request(response)
    except Exception as Argument:

      logging.basicConfig(filename=self.log_file_name, level=logging.DEBUG)
      logging.error("send_message exception", exc_info=True)
      result= {"status_code": 404, "error": Argument}
    return result

  def send_photo(self, image_url= None, file_id= None, file= None):
    try:
      result = {}
      if self.chat_id is None: return {"status_code": 400, "error": "Please send a message to bot"}
      if image_url is None and file_id is None and file is None: return {"status_code": 400, "error": "Please pass at least one file osurce"}

      photo = image_url
      if not photo: photo = file_id
      if not photo: photo = file 

      params = {'chat_id': self.chat_id, 'photo': photo, 'parse_mode': 'HTML'}
      method = SEND_PHOTO
      response = requests.post(self.bot_url + method, params)
      result = self._handle_request(response)
    except Exception as Argument:

      logging.basicConfig(filename=self.log_file_name, level=logging.DEBUG)
      logging.error("send_message exception", exc_info=True)
      result= {"status_code": 404, "error": Argument}
    return result

  def get_response(self, limit: int = 1):
    try:
      get_result =  self._get_all_responses()
      result = self._handle_request(get_result)
      if result.get("status_code") is not 200: return result
      if len(get_result) > 0:
        data = []
        if limit < 1: raise ValueError
        limit = min(len(get_result), limit)
        # start_index = -1 * limit
        # for i in range(limit):
        #   index  = start_index + i
        #   last_update = get_result[index]
        #   response_text = get_object_by_keys(last_update, ["message", "text"])
        #   response_user_info = get_object_by_keys(last_update, ["message", "from"])
        #   data.append({'response':{
        #     'text': response_text,
        #     'user_info': response_user_info
        #   }})
        start_index = len(get_result) - limit
        result['data'] = (result.get('data') or [])[start_index:]
      else:
        result = {"status_code": 400, "error": "Not found response"}
    except ValueError:
      result= {"status_code": 404, "error": "Please enter number over 0"}
    except Exception as Argument:
      logging.basicConfig(filename=self.log_file_name, level=logging.DEBUG)
      logging.error("send_message exception", exc_info=True)
      result= {"status_code": 404, "error": Argument}
    return result

  def _get_all_responses(self, offset=0, timeout=10):
    result_json = []
    try:
      method = GET_UPDATE
      params = {'timeout': timeout, 'offset': offset}
      resp = requests.get(self.bot_url + method, params)
      result_json = get_object_by_keys(resp.json(),'result')
    except Exception as Argument:
      logging.basicConfig(filename=self.log_file_name, encoding='utf-8', level=logging.DEBUG)
      logging.error("get_all_responses exception", exc_info=True)

    return result_json

  def _get_chat_id(self):
    # if self.chat_id: return self.chat_id
    select_result = self._select_chat_id()
    chat_id = None
    if select_result: 
      return select_result
    else: 
      results = self._get_all_responses()
      if results: 
        chat_id = get_object_by_keys(results[0], ["message", "chat", "id"])
        if chat_id: 
          first_name = get_object_by_keys(results[0], ["message", "chat", "id"])
          last_name = get_object_by_keys(results[0], ["message", "chat", "id"])
          self._insert_chat_id(chat_id)
          self._update_chat_user_info(first_name, last_name, chat_id)

      return chat_id

  def _create_table(self):
    self.cursor.execute("CREATE TABLE IF NOT EXISTS telegram_chat_infos (chat_id INTEGER NOT NULL PRIMARY KEY, bot_token TEXT NOT NULL, first_name VARCHAR, last_name VARCHAR)")
    self.connection.commit()

  def _insert_chat_id(self, chat_id):
    self.cursor.execute(f"INSERT OR IGNORE INTO telegram_chat_infos (chat_id, bot_token) VALUES ({chat_id}, '{self.bot_token}')")
    self.connection.commit()

  def _update_chat_user_info(self, first_name, last_name, chat_id):
    self.cursor.execute(f"UPDATE telegram_chat_infos SET first_name = '{first_name}', last_name= '{last_name}' WHERE chat_id = {chat_id}")
    self.connection.commit()  

  def _select_chat_id(self):
    rows = self.cursor.execute(f"SELECT chat_id from telegram_chat_infos where bot_token = '{self.bot_token}'").fetchall()
    if not rows: return None
    row = rows[0]
    if row: return list(row)[0]
    return None

  def _handle_request(self, response):
    if not response: return {"status_code": 404, "error": "missing response"}
    response = response.json()
    if not get_object_by_keys(response,'ok'): return {"status_code": response["error_code"], "error": response["description"]}
    result = get_object_by_keys(response,'result')
    user_info = get_object_by_keys(result,['message', 'from'])
    date = user_info = result.get('date')
    text = user_info = result.get('text')

    return {
      "status_code": 200,
      "data": [{
          user_info: user_info,
          date: date,
          text: text,
        }]
    }






    


